@@ -1,13 +1,14 @@
 //! Functions using all supported type patterns.
 
+use crate::patterns::success_enum::FFIError;
 use crate::types::{
-    ambiguous1, ambiguous2, common, some_foreign_type, CallbackFFISlice, Callbacku8u8, Context, Empty, EnumDocumented, FFIError, Generic, Opaque, Phantom,
-    SomeForeignType, StructDocumented, UseAsciiStringPattern, Vec3f32,
+    ambiguous1, ambiguous2, common, some_foreign_type, Callbacku8u8, Empty, EnumDocumented, Generic, Generic2, Generic3, Opaque, Phantom, SomeForeignType,
+    StructDocumented, Tupled, Vec3f32,
 };
 use interoptopus::ffi_function;
-use interoptopus::patterns::ascii_pointer::AsciiPointer;
-use interoptopus::patterns::slice::FFISlice;
+use interoptopus::patterns::success_enum::panics_and_errors_to_ffi_enum;
 use std::ptr::null;
+use std::time::Duration;
 
 #[ffi_function]
 #[no_mangle]
@@ -21,55 +22,67 @@ pub extern "C" fn primitive_void2() -> () {}
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_bool(x: bool) -> bool {
-    x
+    !x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_u8(x: u8) -> u8 {
-    x
+    u8::MAX - x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_u16(x: u16) -> u16 {
-    x
+    u16::MAX - x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_u32(x: u32) -> u32 {
-    x
+    u32::MAX - x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_u64(x: u64) -> u64 {
-    x
+    u64::MAX - x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_i8(x: i8) -> i8 {
-    x
+    -x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_i16(x: i16) -> i16 {
-    x
+    -x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_i32(x: i32) -> i32 {
-    x
+    -x
 }
 
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn primitive_i64(x: i64) -> i64 {
-    x
+    -x
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn many_args_5(x0: i64, x1: i64, x2: i64, x3: i64, x4: i64) -> i64 {
+    x0 + x1 + x2 + x3 + x4
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn many_args_10(x0: i64, x1: i64, x2: i64, x3: i64, x4: i64, x5: i64, x6: i64, x7: i64, x8: i64, x9: i64) -> i64 {
+    x0 + x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9
 }
 
 #[ffi_function]
@@ -84,45 +97,57 @@ pub extern "C" fn ptr_ptr(x: *const *const i64) -> *const *const i64 {
     x
 }
 
+/// # Safety
+///
+/// Parameter x must point to valid data.
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn ptr_mut(x: *mut i64) -> *mut i64 {
+#[allow(unused_unsafe)]
+pub unsafe extern "C" fn ptr_mut(x: *mut i64) -> *mut i64 {
+    unsafe { *x = -*x };
     x
 }
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn ptr_simple(x: &i64) -> &i64 {
+pub extern "C" fn ref_simple(x: &i64) -> &i64 {
     x
 }
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn ptr_simple_mut(x: &mut i64) -> &mut i64 {
+pub extern "C" fn ref_mut_simple(x: &mut i64) -> &mut i64 {
+    *x = -*x;
     x
 }
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn ptr_option(x: Option<&i64>) -> bool {
+pub extern "C" fn ref_option(x: Option<&i64>) -> bool {
     x.is_some()
 }
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn ptr_option_mut(x: Option<&mut i64>) -> bool {
+pub extern "C" fn ref_mut_option(x: Option<&mut i64>) -> bool {
     x.is_some()
 }
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn complex_1(_a: Vec3f32, _b: Option<&Empty>) -> FFIError {
+pub extern "C" fn tupled(x: Tupled) -> Tupled {
+    Tupled(x.0 * 2)
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn complex_args_1(_a: Vec3f32, _b: Option<&Empty>) -> FFIError {
     FFIError::Ok
 }
 
 #[ffi_function(surrogates(_cmplx = "some_foreign_type"))]
 #[no_mangle]
-pub extern "C" fn complex_2(_cmplx: SomeForeignType) -> *const Opaque {
+pub extern "C" fn complex_args_2(_cmplx: SomeForeignType) -> *const Opaque {
     null()
 }
 
@@ -134,10 +159,28 @@ pub extern "C" fn callback(callback: Callbacku8u8, value: u8) -> u8 {
 
 #[ffi_function]
 #[no_mangle]
-pub extern "C" fn generic(x: Generic<u32>, _y: Phantom<u8>) -> u32 {
+pub extern "C" fn generic_1(x: Generic<u32>, _y: Phantom<u8>) -> u32 {
     *x.x
 }
 
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn generic_2(x: Generic<u8>, _y: Phantom<u8>) -> u8 {
+    *x.x
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn generic_3(x: &Generic2<u8>) -> u8 {
+    x.x
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn generic_4(x: &Generic3<u8>) -> u8 {
+    x.x
+}
+
 /// This function has documentation.
 #[ffi_function]
 #[no_mangle]
@@ -145,81 +188,6 @@ pub extern "C" fn documented(_x: StructDocumented) -> EnumDocumented {
     EnumDocumented::A
 }
 
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_ascii_pointer(x: AsciiPointer, y: UseAsciiStringPattern) -> u8 {
-    let _ = dbg!(x.as_str());
-    let _ = dbg!(y.ascii_string.as_str().unwrap());
-    0
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_class_create(context_ptr: Option<&mut *mut Context>, value: u32) -> FFIError {
-    let the_box = Box::new(Context { some_field: value });
-
-    match context_ptr {
-        None => FFIError::Null,
-        Some(c) => {
-            *c = Box::into_raw(the_box);
-            FFIError::Ok
-        }
-    }
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_ffi_slice(ffi_slice: FFISlice<u32>) -> u32 {
-    ffi_slice.as_slice().unwrap_or(&[]).len() as u32
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_ffi_slice_delegate(callback: CallbackFFISlice) -> u8 {
-    callback.call(FFISlice::from_slice(&[1, 2, 3]))
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_class_method(context: Option<&mut Context>) -> u32 {
-    match context {
-        None => 0,
-        Some(c) => {
-            dbg!(c.some_field);
-            c.some_field *= 2;
-            c.some_field
-        }
-    }
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_class_method_success_enum_ok(_context: Option<&mut Context>) -> FFIError {
-    FFIError::Ok
-}
-
-#[ffi_function]
-#[no_mangle]
-pub extern "C" fn pattern_class_method_success_enum_fail(_context: Option<&mut Context>) -> FFIError {
-    FFIError::Fail
-}
-
-/// # Safety
-///
-/// This function may only be called with a context returned by a succeeding `pattern_class_create`.
-#[ffi_function]
-#[no_mangle]
-#[allow(unused_unsafe)]
-pub unsafe extern "C" fn pattern_class_destroy(context_ptr: Option<&mut *mut Context>) -> FFIError {
-    match context_ptr {
-        None => FFIError::Null,
-        Some(c) => {
-            unsafe { Box::from_raw(*c) };
-            FFIError::Ok
-        }
-    }
-}
-
 #[ffi_function]
 #[no_mangle]
 pub extern "C" fn ambiguous_1(x: ambiguous1::Vec) -> ambiguous1::Vec {
@@ -243,3 +211,15 @@ pub extern "C" fn ambiguous_3(x: ambiguous1::Vec, y: ambiguous2::Vec) -> bool {
 pub extern "C" fn namespaced_type(x: common::Vec) -> common::Vec {
     x
 }
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn panics() -> FFIError {
+    panics_and_errors_to_ffi_enum(|| panic!("Oh no"))
+}
+
+#[ffi_function]
+#[no_mangle]
+pub extern "C" fn sleep(millis: u64) {
+    std::thread::sleep(Duration::from_millis(millis));
+}
