@@ -1,12 +1,10 @@
 //! All supported type patterns.
 
-use interoptopus::ffi_type;
 use interoptopus::lang::c::{CType, CompositeType, Field, PrimitiveType};
 use interoptopus::lang::rust::CTypeInfo;
 use interoptopus::patterns::ascii_pointer::AsciiPointer;
-use interoptopus::patterns::callbacks::CallbackXY;
 use interoptopus::patterns::slice::FFISlice;
-use interoptopus::patterns::success_enum::Success;
+use interoptopus::{ffi_type, pattern_callback};
 use std::marker::PhantomData;
 
 // Let's assume we can't implement `CTypeInfo` for this.
@@ -30,6 +28,10 @@ pub struct Opaque {
     _internal: *const Vec3f32,
 }
 
+#[ffi_type]
+#[repr(C)]
+pub struct Tupled(pub u8);
+
 #[ffi_type]
 #[repr(C)]
 pub struct Generic<'a, T>
@@ -40,6 +42,21 @@ where
     pub x: &'a T,
 }
 
+#[ffi_type(opaque)]
+#[repr(C)]
+pub struct Generic2<T>
+where
+    T: CTypeInfo,
+{
+    pub x: T,
+}
+
+#[ffi_type(opaque, name = "Generic3")]
+#[repr(C)]
+pub struct Generic3<T> {
+    pub x: T,
+}
+
 #[ffi_type(skip(p))]
 #[repr(C)]
 pub struct Phantom<'a, T>
@@ -53,6 +70,7 @@ where
 
 #[ffi_type]
 #[repr(C)]
+#[derive(Copy, Clone, Default)]
 pub struct Vec3f32 {
     pub x: f32,
     pub y: f32,
@@ -66,18 +84,6 @@ pub struct Container {
     pub foreign2: SomeForeignType,
 }
 
-#[ffi_type(patterns(success_enum))]
-#[repr(C)]
-pub enum FFIError {
-    Ok = 0,
-    Null = 100,
-    Fail = 200,
-}
-
-impl Success for FFIError {
-    const SUCCESS: Self = Self::Ok;
-}
-
 /// Documented enum.
 #[ffi_type]
 #[repr(C)]
@@ -102,8 +108,7 @@ pub struct UseAsciiStringPattern<'a> {
     pub ascii_string: AsciiPointer<'a>,
 }
 
-// Used for the `context_init(**Context)`, `context_use(*Context)`,
-// `context_destroy(**Context)` pattern that acts like constructors / methods / destructors.
+/// This can also be used for the `class` pattern.
 #[ffi_type(opaque)]
 #[repr(C)]
 pub struct Context {
@@ -116,7 +121,7 @@ pub type Callbacku8u8 = extern "C" fn(u8) -> u8;
 // This does not work since we can't express the for<'x> bounds in our CTypeInfo implementation.
 // pub type CallbackFFISlice = extern "C" fn(FFISlice<u8>) -> u8;
 
-pub type CallbackFFISlice<'a> = CallbackXY<FFISlice<'a, u8>, u8>;
+pattern_callback!(CallbackFFISlice(slice: FFISlice<u8>) -> u8);
 
 pub mod ambiguous1 {
     use interoptopus::ffi_type;
@@ -164,3 +169,29 @@ pub mod common {
         pub z: f64,
     }
 }
+
+pub mod associated_types {
+    use interoptopus::ffi_type;
+
+    pub trait Helper {
+        type X;
+    }
+
+    #[ffi_type]
+    #[repr(C)]
+    pub struct Chicken(u8);
+
+    #[ffi_type]
+    #[repr(C)]
+    pub struct Cow(u16);
+
+    impl Helper for Chicken {
+        type X = Cow;
+    }
+
+    #[ffi_type]
+    #[repr(C)]
+    pub struct FieldsViaAssociatedType {
+        pub x: <Chicken as Helper>::X,
+    }
+}
